// GrandBrownTown.cuh
//
// Terrance Howard <heyterrance@gmail.com>

#pragma once

__device__
Vector3 step(Vector3 r0, float kTlocal, Vector3 force, float diffusion,
						 float timestep, BaseGrid *sys, Random *randoGen, int num);

__device__
Vector3 step(Vector3 r0, float kTlocal, Vector3 force, float diffusion,
						 Vector3 diffGrad, float timestep, BaseGrid *sys,
						 Random *randoGen, int num);

__global__
void updateKernel(Vector3 pos[], Vector3 forceInternal[],
									int type[], BrownianParticleType* part[],
									float kT, BaseGrid* kTGrid,
									float electricField, int tGridLength,
									float timestep, int num, BaseGrid* sys,
									Random* randoGen, int numReplicas) {
	// Calculate this thread's ID
	const int idx = blockIdx.x * blockDim.x + threadIdx.x;

	// Loop over ALL particles in ALL replicas
	if (idx < num * numReplicas) {
		const int t = type[idx];
		Vector3& p = pos[idx];

		const BrownianParticleType& pt = *part[t];

		// Compute external force
		Vector3 forceExternal = Vector3(0.0f, 0.0f, pt.charge * electricField);

		// Compute PMF
		ForceEnergy fe = pt.pmf->interpolateForceD(p);

#ifndef FORCEGRIDOFF
		// Add a force defined via 3D FORCE maps (not 3D potential maps)
		if (pt.forceXGrid != NULL) fe.f.x += pt.forceXGrid->interpolatePotential(p);
		if (pt.forceYGrid != NULL) fe.f.y += pt.forceYGrid->interpolatePotential(p);
		if (pt.forceZGrid != NULL) fe.f.z += pt.forceZGrid->interpolatePotential(p);		
#endif

		// Compute total force:
		//	  Internal:  interaction between particles
		//	  External:  electric field (now this is basically a constant vector)
		//	  forceGrid: ADD force due to PMF or other potentials defined in 3D space
		Vector3 force = forceInternal[idx] + forceExternal + fe.f;

		if (idx == 0)
			forceInternal[idx] = force; // write it back out for force0 in run()

		// Get local kT value
		float kTlocal = (tGridLength == 0) ? kT : kTGrid->interpolatePotential(p);

		// Update the particle's position using the calculated values for time, force, etc.
		if (pt.diffusionGrid == NULL) {
			p = step(p, kTlocal, force, pt.diffusion, timestep, sys, randoGen, num);
		} else {
			float diffusion = pt.diffusionGrid->interpolatePotential(p);
			Vector3 diffGrad = (pt.diffusionGrid->interpolateForceD(p)).f;
			p = step(p, kTlocal, force, diffusion, -diffGrad, timestep, sys, randoGen, num);
		}
	}
}

__device__
Vector3 step(Vector3 r0, float kTlocal, Vector3 force, float diffusion,
						 float timestep, BaseGrid *sys, Random *randoGen, int num) {
	const int idx = blockIdx.x * blockDim.x + threadIdx.x;
	Vector3 rando = randoGen->gaussian_vector(idx, num);
	Vector3 r = r0 + (force * diffusion * timestep / kTlocal) + (sqrtf(2.0f * diffusion * timestep) * rando);
	Vector3 l = sys->getInverseBasis().transform(r - sys->getOrigin());
	l.x = sys->wrapFloat(l.x, sys->getNx());
	l.y = sys->wrapFloat(l.y, sys->getNy());
	l.z = sys->wrapFloat(l.z, sys->getNz());
	return sys->getBasis().transform(l) + sys->getOrigin();
}

__device__
Vector3 step(Vector3 r0, float kTlocal, Vector3 force, float diffusion,
						 Vector3 diffGrad, float timestep, BaseGrid *sys,
						 Random *randoGen, int num) {
	const int idx = blockIdx.x * blockDim.x + threadIdx.x;
	Vector3 rando = randoGen->gaussian_vector(idx, num);

	Vector3 r = r0 + (diffusion * timestep / kTlocal) * force
							+ timestep * diffGrad
							+ sqrtf(2.0f * diffusion * timestep) * rando;

	Vector3 l = sys->getInverseBasis().transform(r - sys->getOrigin());
	l.x = sys->wrapFloat(l.x, sys->getNx());
 	l.y = sys->wrapFloat(l.y, sys->getNy());
 	l.z = sys->wrapFloat(l.z, sys->getNz());

	return sys->getBasis().transform(l) + sys->getOrigin();
}

__global__ void devicePrint(RigidBodyType* rb[]) {
	// printf("Device printing\n");
	int i = 0;
	printf("RigidBodyType %d: numGrids = %d\n", i, rb[i]->numPotGrids);
	printf("  RigidBodyType %d: potGrid: %p\n", i, rb[i]->rawPotentialGrids);
	int j = 0;
	printf("  RigidBodyType %d: potGrid[%d]: %p\n", i, j, &(rb[i]->rawPotentialGrids[j]));
	printf("  RigidBodyType %d: potGrid[%d] size: %d\n", i, j, rb[i]->rawPotentialGrids[j].getSize());
	// BaseGrid g = rb[i]->rawPotentialGrids[j];
	// for (int k = 0; k < rb[i]->rawPotentialGrids[j].size(); k++)
	for (int k = 0; k < rb[i]->rawPotentialGrids[j].getSize(); k++)
		printf("    rbType_d[%d]->potGrid[%d].val[%d]: %g\n",
					 i, j, k, rb[i]->rawPotentialGrids[j].val[k]);
	// i, j, k, rb[i]->rawPotentialGrids[j]).val[k];
	
}

// __global__ void devicePrint(RigidBodyType* rb[]) {
// 	// printf("Device printing\n");
// 	int i = 0;
// 	printf("RigidBodyType %d: numGrids = %d\n", i, rb[i]->numPotGrids);
// 	printf("RigidBodyType %d: potGrid: %p\n", i, rb[i]->rawPotentialGrids);
// 	int j = 0;
// 	printf("RigidBodyType %d: potGrid[%d]: %p\n", i, &(rb[i]->rawPotentialGrids[j]));
// 	BaseGrid g = rb[i]->rawPotentialGrids[j];
// 	// for (int k = 0; k < rb[i]->rawPotentialGrids[j].size(); k++)
// 	for (int k = 0; k < g->getSize(); k++)
// 		printf("rbType_d[%d]->potGrid[%d].val[%d]: %g\n",
// 					 i, j, k, g.val[k]);
// 	// i, j, k, rb[i]->rawPotentialGrids[j]).val[k];
	
// }
